//******************************************************************
//* VINCULUM THUMB DRIVE CODE
//* LAST UPDATED 11/23/2008
//* FIRMWARE BY FRED EADY FOR NUTS AND VOLTS MAGAZINE
//* CHANGES/ADDITIONS
//******************************************************************
//* THIS CODE PROVIDES A BASIS FOR USING AND WRITING THE VINCULUM
//* COMMANDS. THE COMPLETE SET OF VINCULUM COMMANDS IS NOT INCLUDED
//* IN THIS CODE SET. 
//* IF YOU HAVE ANY QUESTIONS OR NEED HELP ON GETTING A PARTICULAR
//* VINCULUM COMMAND TO WORK, PLEASE SEND AN EMAIL TO fred@edtp.com.
//* UPDATES WILL BE POSTED ON THE EDTP WEB SITE: www.edtp.com
//*
//* THIS CODE WRITEN USING HI-TECH PICC18 STD OR PRO
#include <pic18.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <htc.h>
#include <stdlib.h>

// Settings for PIC18LF2620
__CONFIG(1,IESODIS & FCMDIS & RCIO);
__CONFIG(2,BORDIS & PWRTDIS & WDTDIS);
__CONFIG(3,MCLREN & LPT1DIS & PBDIGITAL);
__CONFIG(4,DEBUGEN & LVPDIS & XINSTDIS & STVREN);
__CONFIG(5,UNPROTECT);
__CONFIG(6,WRTEN);
__IDLOC(1027);

//******************************************************************
//*	FUNCTION PROTOTYPES
//******************************************************************
void init(void);
void cleaner(void);
void putch(char c);
char recvchar(void);
char sendchar(char usart_data);
char CharInQueue(void);
void init_EUSART(void);
void reset_tmr(char timer);
void cmd_echo(char cmd);
char sync(void);
char query_disk(void);
char open_file(void);
char write_file(void);
char close_file(void);
void vreset(void);
//******************************************************************
//*	INTERRUPT DEFINITIONS
//******************************************************************
#define		disable_RXint	  	RCIE = 0
#define		enable_RXint	  	RCIE = 1
#define		disable_TMR1int  	TMR1IE = 0
#define		enable_TMR1int	 	TMR1IE = 1

#define		enable_GLOBALint  	IPEN = 0;				\
							  	GIE = 1;				\
							  	PEIE = 1;
#define		disable_GLOBALint 	IPEN = 0;				\
							  	GIE = 0;				\
							  	PEIE = 0;
#define		TIMER1ON			TMR1ON = 1
#define		TIMER1OFF			TMR1ON = 0

#define mdelay1(msecdelay)		TIMER1OFF;					\
								TMR1IF = 0;					\
								imsecs1 = 0;				\
								tmsecs1 = 0;				\
								TIMER1ON;					\
								while(tmsecs1 < msecdelay);
								
#define	sdelay1(secdelay)		TIMER1OFF;					\
								TMR1IF = 0;					\
								imsecs1 = 0;				\
								tmsecs1 = 0;				\
								isecs1 = 0;					\
								tsecs1 = 0;					\
								TIMER1ON;					\
								while(tsecs1 < secdelay);
//******************************************************************
//*	GENERAL DEFINITONS
//******************************************************************
#define	OFF	0
#define ON	1

typedef struct{
		char E:1;
		char e:1;
		char buf_ovrf:1;
	      }SSTATS;

SSTATS status;

enum {														
	waiting_for_work,											
	disk_ready,
	file_create,
	file_write,
	file_end,
	waiting_for_remove
};
//******************************************************************
//*	PIC18F2620 RAM Definitions
//******************************************************************
unsigned int tmsecs1,imsecs1,tsecs1;
char isecs1;
char sync_loops,current_mode;
char bytein[512];
unsigned int bytein_ptr;
//******************************************************************
//*	EUSART BUFFER DEFINITIONS
//******************************************************************
//1,2,4,8,16,32,64,128 or 256 bytes 
#define EUSART_RX_BUFFER_SIZE  256     
#define EUSART_RX_BUFFER_MASK ( EUSART_RX_BUFFER_SIZE - 1 )
// 1,2,4,8,16,32,64,128 or 256 bytes 
#define EUSART_TX_BUFFER_SIZE  256      
#define EUSART_TX_BUFFER_MASK ( EUSART_TX_BUFFER_SIZE - 1 )
char EUSART_RxBuf[EUSART_RX_BUFFER_SIZE],EUSART_TxBuf[EUSART_TX_BUFFER_SIZE];
char EUSART_TxHead,EUSART_TxTail,EUSART_RxHead,EUSART_RxTail;
//******************************************************************
//*	VDIP2 I/O DEFINITIONS
//******************************************************************
#define	pRTS		LATC0
#define pCTS		LATC1
#define pRST		LATC2
//******************************************************************
//*	LED DEFINITIONS
//******************************************************************
#define LED3		LATC3
#define LED2		LATC4
#define LED1		LATC5

	
void putch(char c)
{
   while(!TRMT) 	// TRMT is set when TSR is empty
      continue;

  TXREG = c;		// load the register
}
//******************************************************************
//*	INTERRUPT HANDLER ROUTINE
//******************************************************************
void interrupt EUSART_TIMER1(void)
{
	  char usart_data,tmphead,tmptail;


	if((TMR1IF && TMR1IE))
	{
		TMR1IF = 0;
		TMR1H = 0xF8;	
		TMR1L = 0x31;

		++tmsecs1;
		if(++imsecs1 == 1000)
		{
			imsecs1 = 0;
			++isecs1;
			++tsecs1;
			LED1 ^= 1;
		}
	}
  if(RCIF)
 {
   usart_data = RCREG;			//read the received usart_data 
   // calculate buffer index 
   tmphead = ( EUSART_RxHead + 1 ) & EUSART_RX_BUFFER_MASK;
   EUSART_RxHead = tmphead;		// store new index 

   if ( tmphead == EUSART_RxTail )
   {
     //ERROR! Receive buffer overflow 
     status.buf_ovrf = 1;
   }
                                
   EUSART_RxBuf[tmphead] = usart_data;   // store received usart_data in buffer 
  }

  
  if(TRMT)
  {
   //check if all usart_data is transmitted  
   if ( EUSART_TxHead != EUSART_TxTail )
   {
     //calculate new buffer index  
     tmptail = ( EUSART_TxTail + 1 ) & EUSART_TX_BUFFER_MASK;
     EUSART_TxTail = tmptail;      	// store new index  

    TXREG = EUSART_TxBuf[tmptail];  // start transmission  
   }
   else
   {
     TXIE = 0;         				// disable TX interrupt  
   }
  }
} 
//******************************************************************
//*	RETRIEVE A CHARACTER FROM THE RECEIVE BUFFER
//******************************************************************
char recvchar(void)
{
  char tmptail;
  // wait for incomming usart_data  
  while ( EUSART_RxHead == EUSART_RxTail );
  // calculate buffer index  
  tmptail = ( EUSART_RxTail + 1 ) & EUSART_RX_BUFFER_MASK;
  //store new index
  EUSART_RxTail = tmptail;        
  //return usart_data
  return EUSART_RxBuf[tmptail];   
}
//******************************************************************
//*	SEND A CHARACTER FROM THE TRANSMIT BUFFER
//******************************************************************
char sendchar(char usart_data)
{
  char tmphead;
                                // calculate buffer index 
  tmphead = ( EUSART_TxHead + 1 ) & EUSART_TX_BUFFER_MASK;
                                // wait for free space in buffer 
  while ( tmphead == EUSART_TxTail );
                                // store usart_data in buffer 
  EUSART_TxBuf[tmphead] = (char)usart_data;
  EUSART_TxHead = tmphead;        // store new index 

  TXIE = 1;            // enable TX interrupt 

  return usart_data;
}
//******************************************************************
//*	CHECK FOR CHARACTER IN RECEIVE BUFFER
//******************************************************************
char CharInQueue(void)
{
  return(EUSART_RxHead != EUSART_RxTail);
}
//******************************************************************
//*	CLEAR THE RECEIVE BUFFER
//******************************************************************
void cleaner(void) //loads receive buffer with 0x00
{
	unsigned int x;
	for(x=0;x<EUSART_RX_BUFFER_SIZE-1;++x)
		EUSART_RxBuf[x] = 0;
	for(x=0;x<EUSART_TX_BUFFER_SIZE-1;++x)
		EUSART_TxBuf[x] = 0;
  	EUSART_RxTail = 0x00;	//flush receive buffer 
  	EUSART_RxHead = 0x00;
  	EUSART_TxTail = 0x00;	//flush transmit buffer
  	EUSART_TxHead = 0x00;
}
//******************************************************************
//*	Init EUSART Function
//******************************************************************
void init_EUSART(void)
{
  SPBRG = 51;	//8MHZ = 51 FOR 9600BAUD
  TRISC7 = 1;	//receive pin
  TRISC6 = 0;	//transmit pin
  TXSTA = 0x04;	//high speed baud rate set BRGH = 1
  RCSTA = 0x80;	//enable serial port and serial port pins
  
  EUSART_RxTail = 0x00;	//flush receive buffer 
  EUSART_RxHead = 0x00;
  EUSART_TxTail = 0x00;	//flush transmit buffer
  EUSART_TxHead = 0x00;
  RCIP = 1;			//receive interrupt = high priority
  TXIP = 1;			//transmit interrupt = high priority
  RCIE = 1;			//enable receive interrupt
  PEIE = 1;			//enable all unmasked peripheral interrupts
  GIE = 1;			//enable all unmasked interrupts
  CREN = 1;			//enable EUSART1 receiver
  TXIE = 0;			//disable EUSART1 transmit interrupt
  TXEN = 1;			//transmitter enabled
}
//******************************************************************
//*	ECHO Command
//*	Send "E" or "e" plus carriage return
//******************************************************************
void cmd_echo(char cmd)
{
	printf("%c\r",cmd);								
}
//******************************************************************
//*	DIR
//******************************************************************
char vDIR(void)
{
	char rc;
	
	rc = 0;
	sync_loops = 0;
	bytein_ptr = 0;
	printf("DIR\r");
	
		while((!CharInQueue()));	//wait for character
		do{
			bytein[bytein_ptr++] = recvchar();
		}while(CharInQueue());	
			
	
		do{
		sendchar(0x0D);	
		++sync_loops;
		bytein_ptr = 0;				//init recv character array pointer
		sendchar(0x0D);				// Send carriage return
		while((!CharInQueue()));	//wait for character
		do{
			bytein[bytein_ptr] = recvchar();
			if(bytein[bytein_ptr] == 'D')
			{
				for(bytein_ptr=1;bytein_ptr<6;++bytein_ptr)
					bytein[bytein_ptr] = recvchar();
			}	
		
		}while(CharInQueue());										
			//check for prompt
		if(bytein_ptr)							
		{
			if ((bytein[0] == 'D') 	&& 
				(bytein[1] == ':') 	&&
				(bytein[2] == 0x5C)	&&
				(bytein[3] == '>') 	&&
				(bytein[4] == 0x0D))				
			{
				rc = 1;
			}
		}	
	}while((sync_loops < 255) && (rc == 0));

	return rc;
}
//******************************************************************
//*	SYNC WITH VINCULUM
//******************************************************************
char sync(void)
{
	char rc;
	
	rc = 0;
	sync_loops = 0;	//init sync loop counter
	status.E = 0;	//clear E received status
	status.e = 0;	//clear e received status
	
	cleaner();	//clear receive buffer

	do{
		cmd_echo('E');				//send E
		sync_loops++;				//increment sync loop counter
		bytein_ptr = 0;				//init recv character array pointer
		while((!CharInQueue()));	//wait for character
		//get 2 bytes
			do{
			bytein[bytein_ptr++] = recvchar();	
		}while(CharInQueue());										
		//check for echoed E and carriage return
		if(bytein_ptr)							
		{
			if ((bytein[0] == 'E') && (bytein[1] == 0x0D))
			{
				status.E = 1;
			}
			else
			{
				mdelay1(100);								
			}
		}	
	}while((status.E == 0) && (sync_loops < 0xFF));

	if (status.E)
	{
		do{
			cmd_echo('e');				//send e
			sync_loops++;				//increment sync loop counter
			bytein_ptr = 0;				//init recv character array pointer
			while((!CharInQueue()));	//wait for character
			//get 2 bytes
				do{
				bytein[bytein_ptr++] = recvchar();	
			}while(CharInQueue());										
			//check for echoed e and carriage return
			if(bytein_ptr)							
			{
				if ((bytein[0] == 'e') && (bytein[1] == 0x0D))
				{
					status.e = 1;
				}
				else
				{
					mdelay1(100);								
				}
			}	
		}while((status.e == 0) && (sync_loops < 0xFF));
	}
	if(status.E && status.e)
		rc = 1;

	return rc;

}
//******************************************************************
//*	CHECK FOR MOUNTED DISK
//******************************************************************
char query_disk(void)										
{
	char rc,x;
	
	rc = 0;
	bytein_ptr = 0;				//init recv character array pointer
	sendchar(0x0D);				// Send carriage return
	while((!CharInQueue()));	//wait for character
	do{
		bytein[bytein_ptr] = recvchar();
		if(bytein[bytein_ptr] == 'D')
		{
				for(bytein_ptr=1;bytein_ptr<6;++bytein_ptr)
					bytein[bytein_ptr] = recvchar();
		}	
		
	}while(CharInQueue());										
	//check for echoed E and carriage return
	if(bytein_ptr)							
	{
		if ((bytein[0] == 'D') 	&& 
			(bytein[1] == ':') 	&&
			(bytein[2] == 0x5C)	&&
			(bytein[3] == '>') 	&&
			(bytein[4] == 0x0D))				
		{
			rc = 1;
		}
	}	
	return rc;
}
//******************************************************************
//*	OPEN FILE FOR WRITE
//******************************************************************
char open_file(void)
{
	char rc;
	
	rc = 0;
	sync_loops = 0;
	
	printf("OPW NV.txt");
	//use sendchar for debugging
/*	
	sendchar('O');									// Send 'O'
	sendchar('P');									// Send 'P'
	sendchar('W');									// Send 'W'
	sendchar(' ');									// Send ' '
	sendchar('N');									// Send 'N'
	sendchar('V');									// Send 'V'
	sendchar('.');									// Send '.'
	sendchar('t');									// Send 't'
	sendchar('x');									// Send 'x'
	sendchar('t');									// Send 't'
*/	
	do{
		sendchar(0x0D);	
		++sync_loops;
		bytein_ptr = 0;				//init recv character array pointer
		sendchar(0x0D);				// Send carriage return
		while((!CharInQueue()));	//wait for character
		do{
			bytein[bytein_ptr] = recvchar();
			if(bytein[bytein_ptr] == 'D')
			{
				for(bytein_ptr=1;bytein_ptr<6;++bytein_ptr)
					bytein[bytein_ptr] = recvchar();
			}	
		
		}while(CharInQueue());										
			//check for prompt
		if(bytein_ptr)							
		{
			if ((bytein[0] == 'D') 	&& 
				(bytein[1] == ':') 	&&
				(bytein[2] == 0x5C)	&&
				(bytein[3] == '>') 	&&
				(bytein[4] == 0x0D))				
			{
				rc = 1;
			}
		}	
	}while((sync_loops < 255) && (rc == 0));

	return rc;

}
//******************************************************************
//*	WRITE FILE
//******************************************************************
char write_file(void)
{
	char rc;
	
	sync_loops = 0;
	rc = 0;
	
	printf("WRF %c%c%c%c\rNUTS AND VOLTS\r\n",0,0,0,14);
	//use sendchar for debugging
/*
	sendchar('W');									// Send 'W'
	sendchar('R');									// Send 'R'
	sendchar('F');									// Send 'F'
	sendchar(' ');									// Send ' '
	sendchar(0x00);									// Send 0x00 - Number of bytes to write MSB
	sendchar(0x00);									// Send 0x00
	sendchar(0x00);									// Send 0x00
	sendchar(0x0E);									// Send 0x0E - Number of  bytes to write LSB
	sendchar(0x0D);									// Send carriage return
	sendchar('N');									// Send 'N'
	sendchar('U');									// Send 'U'
	sendchar('T');									// Send 'T'
	sendchar('S');									// Send 'S'
	sendchar(' ');									// Send ' '
	sendchar('A');									// Send 'A'
	sendchar('N');									// Send 'N'
	sendchar('D');									// Send 'D'
	sendchar(' ');									// Send ' '
	sendchar('V');									// Send 'V'
	sendchar('O');									// Send 'O'
	sendchar('L');									// Send 'L'
	sendchar('T');									// Send 'T'
	sendchar('S');									// Send 'S'
	sendchar(0x0D);									// Send carriage return
	sendchar(0x0A);									// Send line feed
	//sendchar(0x0D);									// Send carriage return
*/	
	do{
		sendchar(0x0D);	
		++sync_loops;
		bytein_ptr = 0;				//init recv character array pointer
		sendchar(0x0D);				// Send carriage return
		while((!CharInQueue()));	//wait for character
		do{
			bytein[bytein_ptr] = recvchar();
			if(bytein[bytein_ptr] == 'D')
			{
				for(bytein_ptr=1;bytein_ptr<6;++bytein_ptr)
					bytein[bytein_ptr] = recvchar();
			}	
		
		}while(CharInQueue());										
			//check for prompt
		if(bytein_ptr)							
		{
			if ((bytein[0] == 'D') 	&& 
				(bytein[1] == ':') 	&&
				(bytein[2] == 0x5C)	&&
				(bytein[3] == '>') 	&&
				(bytein[4] == 0x0D))				
			{
				rc = 1;
			}
		}	
	}while((sync_loops < 255) && (rc == 0));

	return rc;

}

//******************************************************************
//*	CLOSE FILE
//******************************************************************
char close_file(void)
{
	char rc;
	
	sync_loops = 0;
	rc = 0;
	
	printf("CLF NV.txt");
	//use sendchar for debugging
/*
	sendchar('C');									// Send 'C'
	sendchar('L');									// Send 'L'
	sendchar('F');									// Send 'F'
	sendchar(' ');									// Send ' '
	sendchar('N');									// Send 'N'
	sendchar('V');									// Send 'V'
	sendchar('.');									// Send '.'
	sendchar('t');									// Send 't'
	sendchar('x');									// Send 'x'
	sendchar('t');									// Send 't'
*/	
	do{
		sendchar(0x0D);	
		++sync_loops;
		bytein_ptr = 0;				//init recv character array pointer
		sendchar(0x0D);				// Send carriage return
		while((!CharInQueue()));	//wait for character
		do{
			bytein[bytein_ptr] = recvchar();
			if(bytein[bytein_ptr] == 'D')
			{
				for(bytein_ptr=1;bytein_ptr<6;++bytein_ptr)
					bytein[bytein_ptr] = recvchar();
			}	
		
		}while(CharInQueue());										
			//check for prompt
		if(bytein_ptr)							
		{
			if ((bytein[0] == 'D') 	&& 
				(bytein[1] == ':') 	&&
				(bytein[2] == 0x5C)	&&
				(bytein[3] == '>') 	&&
				(bytein[4] == 0x0D))				
			{
				rc = 1;
			}
		}	
	}while((sync_loops < 255) && (rc == 0));

	return rc;

}
//******************************************************************
//*	VDIP2 RESET
//******************************************************************
void vreset(void)
{
	pRST = 0;
	mdelay1(100);
	pRST = 1;
}
//******************************************************************
//*	INITIALIZE ROUTINE
//******************************************************************
void init()
{
//******************************************************************
//*	INITIALIZE CLOCK AND I/O PORTS
//******************************************************************
	OSCCON = 0x70;		//set for 8 MHz operation
	PLLEN = 0;			//no PLL
	TRISA = 0b01111111;
	TRISB = 0b11111111;
	TRISC = 0b10000001;	//RTS is input
//******************************************************************
//*	CONFIGURE A2D AND COMPARATORS
//******************************************************************
	ADCON1 = 0b00001111;	//all digital inputs
	ADON = 0;
	CMCON = 0x07;
//******************************************************************
//*	INITIALIZE EUSART
//******************************************************************
	init_EUSART();
//******************************************************************
//*	CONFIGURE AND START TIMER1
//* SET TO OVERFLOW EVERY 1mS
//******************************************************************
	//F831 = 8MHz
	//E0C1 = 32MHz
	TIMER1OFF;
	T1CON = 0b00000000;
	TMR1H = 0xF8;
	TMR1L = 0x31;
	TIMER1ON;
//******************************************************************
//*	CONFIGURE EXTERNAL INTERRUPTS
//******************************************************************
	enable_TMR1int;
    enable_GLOBALint;
//******************************************************************
//*	INITIALIZE I/O
//******************************************************************
    pCTS = 0;
    pRST = 1;
    LED2 = OFF;
    LED3 = OFF;
    current_mode = waiting_for_work;								
	sdelay1(1);											
}
//******************************************************************
//*	MAIN MAIN MAIN MAIN MAIN
//******************************************************************
void main()
{
	init();
	cleaner();
	vreset();
	sync();
	sendchar(0x0D);
	vDIR();
	do{
		++sync_loops;
	}while(1);	
		
	if (sync())											// Synchronise to Vinculum
	{
		do{								// Main state machine - do not return from here
			switch (current_mode)
			{
				case waiting_for_work:
					if (query_disk())		// Check for disk
					{
						current_mode = disk_ready;	// Update state to indicate disk found
					}
				break;
	
				case disk_ready:
					if(open_file())						// Send open file for write command	(OPW) - file name "hello.txt"
					{
						current_mode = file_write;			// Update state to indicate file has been created and can be written
					}
					else
					{
						current_mode = waiting_for_work;
					}
					break;

				case file_write:
					if(write_file())					// Send write file command (WRF) - write "Hello World!"
					{
						current_mode = file_end;			// Update state to indicate file has been written and can now be closed
					}
					else
					{
						current_mode = waiting_for_work;
					}
					break;

				case file_end:
					if(close_file())						// Send close file command (CLF)
					{
						current_mode = waiting_for_remove;		// Update state to indicate witing for disk to be removed
					}
					else
					{
						current_mode = waiting_for_work;
					}
					break;

				case waiting_for_remove:
					LED2 = 1;
					break;
				
				default:
					current_mode = waiting_for_work;				// Default to idle state
			}

		}while(1);

	}

}

